/*
 * Modem control
 *
 * Copyright (C) 2000-2002, Nam SungHyun and various contributors
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#include "config.h"

#include <termios.h>
#include <fcntl.h>	/* access() */
#include <sys/ioctl.h>	/* TIOCMGET */
#include <sys/stat.h>	/* stat() */
#include <errno.h>
#include <signal.h>	/* kill() */
#include <ctype.h>	/* isdigit() */

#include "pcMain.h"
#include "pcTerm.h"	/* Term */

/* Definitions {{{1 */

#define MODEM_CALL_PREFIX	 "atdt"
#define MODEM_CALL_SUFFIX	 "\r"
#define MODEM_DIAL_CANCEL_STR	 "\r"
#define MODEM_DIAL_WAIT_INTERVAL (30 * 1000)

/* Global variables/functions {{{1 */

/*  Ͽ о phone info  */
GSList *ModemInfoList = NULL;
ModemInfoType *ModemInfo = NULL;

guint32 IdleGuardInterval;
char *IdleGuardString;
gboolean DoLocalEcho;
gboolean DoRemoteEcho;
gboolean DoLock;
guint32 IdleGapBetweenChar;
guint32 IdleGapBetweenLine;

guint32 ModemMaxHistory;
guint32 ModemHistoryBufSize;

/* Local function prototypes {{{1 */

static void modemInfoFree(void);
static void modemAddHistory(const guchar *s, int len);
static void modemHistoryDelete(void);
static void modemHistorySend(GtkCList *clist);
static void modemHistorySendCR(GtkCList *clist);

/* Local variables {{{1 */

static struct termios origTC;
static gboolean ModemInitialized = FALSE;
static gboolean ModemConnected = FALSE;
static gboolean ModemLocked = FALSE;
static gboolean ModemHangupInProgress = FALSE;
static gboolean ModemDirectConnect = FALSE;
static guint ModemDialTimeoutTag = 0;
static gboolean ModemDialInProgress = FALSE;

static ModemInfoType DfltModemInfo = {
    NULL, NULL, 0, NULL, 0, 0, 0, NULL, NULL, NULL, NULL
};

static char **modemHistoryBuf = NULL;
static guint modemHistoryIdx, modemHistoryPos;
static gboolean modemHistoryWrapped;

/* ModemSetSpeed() {{{1 */
gboolean
ModemSetSpeed(guint32 modemSpeed)
{
    long speed;
    struct termios tty;

    if (UseModem)
    {
	tcgetattr(ReadFD, &tty);

	switch (modemSpeed)
	{
	    case 0: speed = B0; break;
	    case 300: speed = B300; break;
	    case 600: speed = B600; break;
	    case 1200: speed = B1200; break;
	    case 2400: speed = B2400; break;
	    case 4800: speed = B4800; break;
	    case 9600: speed = B9600; break;
	    case 19200: speed = B19200; break;
	    case 38400: speed = B38400; break;
	    case 57600: speed = B57600; break;
	    case 115200: speed = B115200; break;
	    default:
		speed = -1;
		break;
	}

	if (speed >= 0)
	{
	    cfsetospeed(&tty, (speed_t) speed);
	    cfsetispeed(&tty, (speed_t) speed);

	    if (tcsetattr(ReadFD, TCSANOW, &tty) == 0)
	    {
		ControlBaudrateChange(modemSpeed);
		ModemInfo->speed = modemSpeed;
		return TRUE;
	    }
	}
    }
    if (ModemInfo)
	ModemInfo->speed = modemSpeed;
    return FALSE;
}

/* ModemGetSpeed() {{{1 */
guint32
ModemGetSpeed(void)
{
    if (ModemInfo)
	return ModemInfo->speed;
    else
	return (guint32) 115200;
}

/* ModemSetParams() {{{1 */
static gboolean
ModemSetParams(ModemInfoType *info)
{
    struct termios tty;

    if (ModemSetSpeed(info->speed) == FALSE)
	return FALSE;

    tcgetattr(ReadFD, &tty);

    switch (info->bits)
    {
	case 5:
	    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS5;
	    break;
	case 6:
	    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS6;
	    break;
	case 7:
	    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS7;
	    break;
	case 8:
	default:
	    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;
	    break;
    }

    tty.c_iflag = IGNBRK;
    tty.c_lflag = 0;
    tty.c_oflag = 0;
    tty.c_cflag |= CLOCAL | CREAD;
    tty.c_cc[VMIN] = 1;
    tty.c_cc[VTIME] = 1;

    tty.c_cflag &= ~(PARENB | PARODD);
    if (!strcasecmp(info->parity, "EVEN"))
	tty.c_cflag |= PARENB;
    else if (!strcasecmp(info->parity, "ODD"))
	tty.c_cflag |= PARODD;

    if (info->useSWF)
	tty.c_iflag |= IXON | IXOFF;
    else
	tty.c_iflag &= ~(IXON | IXOFF | IXANY);

    if (info->useHWF)
	tty.c_cflag |= CRTSCTS;
    else
	tty.c_cflag &= ~CRTSCTS;

    return (tcsetattr(ReadFD, TCSANOW, &tty) == 0);
}

/* ModemResetTTY() {{{1 */
int
ModemResetTTY(void)
{
    struct termios tty;

    if (ModemInfo && ReadFD >= 0)
    {
	tcgetattr(ReadFD, &tty);
	tcsetattr(ReadFD, TCSAFLUSH, &tty);
	return ModemSetParams(ModemInfo);
    }
    return -1;
}

/* ModemGetConnectionStatus() {{{1 */
gboolean
ModemGetConnectionStatus(void)
{
    unsigned int v;

    if (!ModemInitialized)
	return FALSE;

    if (ModemDirectConnect)
    {
	if (ModemHangupInProgress)
	    return FALSE;
	else
	    return TRUE;
    }

    v = 0;
    if (ReadFD != -1)
	ioctl(ReadFD, TIOCMGET, &v);
    return (v & TIOCM_CAR) ? TRUE : FALSE;
}

/* ModemRestore() {{{1 */
static void
ModemRestore(void)
{
    if (ModemInitialized)
    {
	tcsetattr(ReadFD, TCSANOW, &origTC);
	if (DoLock)
	    ModemUnlock();
    }
}

/* ModemDtrToggle() {{{1 */
static void
ModemDtrToggle(int fd)
{
    struct termios tty, old;

    /* POSIX: set baudrate to 0 and back */
    tcgetattr(fd, &tty);
    memcpy(&old, &tty, sizeof(old));
    cfsetospeed(&tty, B0);
    cfsetispeed(&tty, B0);
    tcsetattr(fd, TCSANOW, &tty);
    sleep(1);
    tcsetattr(fd, TCSANOW, &old);
}

/* ModemConnectionCheck() {{{1 */
int
ModemConnectionCheck(gpointer dummy)
{
    guint32 currTime;
    GSList *slist;
    guint32 u32;
    TermType *term;
    static guint32 connectTime = 0;
    static gint32 tick;

    if (ModemConnected != ModemGetConnectionStatus())
    {
	ModemConnected = !ModemConnected;
	if (ModemConnected)
	{
	    tick = -1;
	    connectTime = gdk_time_get();

	    /* logging */
	    if (LogFile)
		OpenLogFile(ModemInfo->name, ModemInfo->number);

	    StatusShowMessage(_("Connected..."));
	    SoundPlay(SoundConnectFile);
	    PC_State = PC_RUN_STATE;
	    PC_ButtonControl(TB_CONNECT_STATE);

	    for (slist = ConnectFuncList; slist;)
	    {
		void (*func) (void) = slist->data;

		/* list remove Ͼ   list  
		 * ϴ. func ڱ ڽ   ֱ ̴.
		 */
		slist = slist->next;

		(*func) ();
	    }
	}
	else
	{
	    LastOutTime += 800000000UL;
	    StatusShowMessage(_("Disconnected..."));
	    SoundPlay(SoundDisconnectFile);
	    SetWindowTitle(NULL);	/* set to default title */
	    PC_State = PC_IDLE_STATE;
	    PC_ButtonControl(TB_IDLE_STATE);
	    ModemInitialized = FALSE;

	    if (ScriptRunning)
		ScriptCancel();

	    for (slist = DisconnectFuncList; slist;)
	    {
		void (*func) (void) = slist->data;

		/* list remove Ͼ   list   ϴ.
		 * func ڱ ڽ   ֱ ̴.
		 */
		slist = slist->next;

		(*func) ();
	    }

	    /* logging */
	    if (LogFile)
		CloseLogFile(tick);

	    if (connectTime > 0 && ModemInfo->logoutScript)
		ScriptRunFromFile(ModemInfo->logoutScript);

	    modemInfoFree();
	    connectTime = 0;
	    close(ReadFD);
	    ReadFD = -1;
	}
    }

    if (ModemConnected)
    {
	currTime = gdk_time_get();

	if (IdleGuardString && *IdleGuardString)
	    if (LastOutTime + (IdleGuardInterval * 1000) < currTime)
		Term->remoteWrite(IdleGuardString, strlen(IdleGuardString));

	if (AR_CheckFlag)
	{
	    AR_CheckFlag = FALSE;
	    term = Term;
	    AutoResponseCheckRun(term->buf
				 + term->col * TERM_LINENUM(term->currY),
				 term->currX);
	}

	u32 = (currTime - connectTime) / 1000;
	if (u32 != (guint32) tick)
	{
	    tick = u32;
	    StatusShowTime(tick);
	}
    }
    return TRUE;

    dummy = 0;	/* to make gcc happy */
}

/* ModemWrite() {{{1 */
int
ModemWrite(const guchar *s, int len)
{
    register int i;
    char c;

    LastOutTime = gdk_time_get();
    if (DoLocalEcho && Term)
    {
	for (i = 0; i < len; i++)
	{
	    if (s[i] != '\r')
		Term->localWrite(Term, s + i, 1);
	    else
	    {
		c = '\n';
		Term->localWrite(Term, s + i, 1);
		Term->localWrite(Term, &c, 1);
	    }
	}
    }

    if (ModemDirectConnect)
    {
	if (modemHistoryBuf && TRxRunning == TRX_IDLE)
	    modemAddHistory(s, len);

	for (i = 0; i < len; i++, s++)
	{
	    write(ReadFD, s, 1);
	    if (IdleGapBetweenLine && (*s == '\n' || *s == '\r'))
		usleep(IdleGapBetweenLine);
	    else if (IdleGapBetweenChar)
		usleep(IdleGapBetweenChar);
	}
    }
    else
	write(ReadFD, s, len);

    ControlLabelLineStsToggle('t');
    return 0;
}

/* ModemLock() {{{1 */
int
ModemLock(void)
{
    register int i, handle;
    int pid;
    char buf[MAX_PATH], lockbuf[128], temp[32];
    struct stat st;

    if (ModemLocked || ModemInfo == NULL)
	return 0;

    strncpy(temp, strrchr(ModemInfo->device, '/') + 1, sizeof(temp) - 1);
    g_snprintf(buf, sizeof(buf), "%s/LCK..%s", MODEM_LOCK_PATH, temp);
    if (stat(buf, &st) == 0)
    {
	if ((handle = open(buf, O_RDONLY)) == -1)
	    goto err_return;
	i = read(handle, &lockbuf, sizeof(lockbuf));
	close(handle);
	if (i > 0)
	{
	    lockbuf[i] = 0;
	    sscanf(lockbuf, "%d", &pid);
	    if (pid > 0 && kill(pid, 0) < 0 && errno == ESRCH)
	    {
		if (unlink(buf) == -1)
		    goto err_return;
	    }
	    else
		goto err_return;
	}
    }

    if ((handle = open(buf, O_CREAT | O_WRONLY | O_EXCL)) == -1)
	goto err_return;
    fchmod(handle, S_IWRITE | S_IREAD | S_IRGRP | S_IROTH);

    g_snprintf(buf, sizeof(buf), "%010d\n", getpid());
    write(handle, buf, strlen(buf));
    close(handle);
    ModemLocked = TRUE;
    return 0;

  err_return:
    ModemLocked = FALSE;
    g_warning("ModemLock: lock fail.");
    return -1;
}

/* ModemUnlock() {{{1 */
int
ModemUnlock(void)
{
    char buf[MAX_PATH], temp[32];

    if (!ModemLocked || !ModemInfo)
	return 0;

    strncpy(temp, strrchr(ModemInfo->device, '/') + 1, sizeof(temp) - 1);
    g_snprintf(buf, sizeof(buf), "%s/LCK..%s", MODEM_LOCK_PATH, temp);
    if (unlink(buf) == -1)
	return -1;
    ModemLocked = FALSE;
    return 0;
}

/* ModemCheckDialNumber() {{{1 */
static gint
ModemCheckDialNumber(const char *number)
{
    /* check number. I allow next format:
     *    9,0333-222-1111
     * , , ̳ʽ, ޸ 밡.
     */
    for (; *number; number++)
	if (!isdigit(*number) && *number != '-' && *number != ',')
	    return ERROR;
    return OK;
}

/* ModemDialTimeoutCB() {{{1 */
static gint
ModemDialTimeoutCB(gpointer data)
{
    ModemDialTimeoutTag = 0;
    StatusShowMessage(_("Modem connect timeout!"));
    ModemWrite(MODEM_DIAL_CANCEL_STR, strlen(MODEM_DIAL_CANCEL_STR));
    PC_Sleep(1000);
    ModemDialInProgress = FALSE;
    return FALSE;
    data = NULL;
}

/* ModemResponseWaitFilter() {{{1 */
static gint
ModemResponseWaitFilter(char *buf, int len)
{
    int i;
    static int log = FALSE;
    static char response[20];
    static char *p = response;

    if (buf == NULL || len <= 0)
    {
	p = response;
	log = FALSE;
	return 0;
    }

    for (i = 0; i < len; i++)
    {
	if (buf[i] == 10)
	{
	    if (log)
	    {
		*p = '\0';
		if (strstr(response, "BUSY"))
		{
		    StatusShowMessage(_("BUSY!"));
		    PC_Sleep(1000);
		    /* TODO: dial next number */
		}
		else if (strstr(response, "NO CARRIER")
			 || strstr(response, "NO DIALTONE")
			 || strstr(response, "OK") || strstr(response, "VOICE"))
		{
		    PC_Sleep(1000);
		    /* TODO: dial next number */
		}
		else
		{
		    ModemConnectionCheck(NULL);
		    /* TODO: hide if there is a dialog */
		}

		gtk_timeout_remove(ModemDialTimeoutTag);
		ModemDialTimeoutTag = 0;
		InputFilterList = g_slist_remove(InputFilterList,
						 ModemResponseWaitFilter);
		p = response;
		log = FALSE;
		ModemDialInProgress = FALSE;
		break;
	    }
	    else
		log = TRUE;
	}
	else
	{
	    if (log && p < (buf + sizeof(buf) - 1))
		*p++ = buf[i];
	}
    }
    return 0;
}

/* ModemDialNumber() {{{1 */
static void
ModemDialNumber(ModemInfoType *info)
{
    char *p;

    if (ModemInitialized)
    {
	StatusShowMessage(_("Modem already initialized. Disconnect first!"));
	return;
    }

    if (ModemCheckDialNumber(info->number) == ERROR)
    {
	StatusShowMessage(_("Invalid dial number '%s'"), info->number);
	return;
    }

    ModemDialInProgress = TRUE;

    ModemSetSpeed(info->speed);
    ModemWrite(MODEM_CALL_PREFIX, strlen(MODEM_CALL_PREFIX));
    for (p = info->number; *p; p++)
    {
	if (*p == ',')
	    PC_Sleep(2000);	/* maybe-NEVER-todo: sleep Ѱ? */
	if (isdigit(*p))
	    ModemWrite(p, 1);
    }
    ModemWrite(MODEM_CALL_SUFFIX, strlen(MODEM_CALL_SUFFIX));

    StatusShowMessage(_("Try to dial '%s'"), info->number);

    if (ModemDialTimeoutTag > 0)
	gtk_timeout_remove(ModemDialTimeoutTag);
    ModemDialTimeoutTag = gtk_timeout_add(MODEM_DIAL_WAIT_INTERVAL,
					  (GtkFunction) ModemDialTimeoutCB,
					  NULL);

    ModemResponseWaitFilter(NULL, -1);	/* init */
    InputFilterList = g_slist_prepend(InputFilterList, ModemResponseWaitFilter);
}

/* modemInfoFree() {{{1 */
static void
modemInfoFree(void)
{
    if (ModemInfo == &DfltModemInfo)
    {
	if (ModemInfo->number)
	{
	    g_free(ModemInfo->number);
	    ModemInfo->number = NULL;
	}
    }
    ModemInfo = NULL;
}

/* InitModem() {{{1 */
static int
InitModem(ModemInfoType *info)
{
    g_return_val_if_fail(info != NULL, -1);

    if (ModemInfo->loginScript)
    {
	ScriptRunFromFile(ModemInfo->loginScript);
	PC_Sleep(100);
    }

    if (DoLock)
    {
	if (ModemLock() < 0)
	{
	    StatusShowMessage(_("Cannot lock device. Modem cannot initialized."));
	    return -1;
	}
    }

    if (access(info->device, R_OK | W_OK))
    {
	FatalExit(NULL, "Cannot access modem device '%s'.", info->device);
	return -1;
    }

    if ((ReadFD = open(info->device, O_RDWR)) < 0)
    {
	FatalExit(NULL, "Cannot open modem device '%s'.", info->device);
	return -1;
    }

    if (tcgetattr(ReadFD, &origTC))
    {
	FatalExit(NULL, "Cannot get status of the modem device '%s'.",
		  info->device);
	return -1;
    }

    if (ModemSetParams(info) == FALSE)
    {
	ScriptCancel();
	FatalExit(NULL, "Modem device '%s' initialization error.",
		  info->device);
	return -1; /* SHOULD NEVER REACH */
    }

    if (info->initString)
	ModemWrite(info->initString, strlen(info->initString));

    ModemConnected = FALSE;

    if (!strncasecmp(info->number, "DIRECT", 6))
	ModemDirectConnect = TRUE;
    else
    {
	ModemDirectConnect = FALSE;
	ModemDialNumber(info);
    }

    gtk_timeout_add(DEFAULT_CONNECT_CHK_INTERVAL, ModemConnectionCheck,
		    NULL);

    ModemHistoryNew();

    ModemInitialized = TRUE;
    return 0;
}

/* ModemExit() {{{1 */
void
ModemExit(void)
{
    GSList *l;
    ModemInfoType *mi;

    for (l = ModemInfoList; l; l = l->next)
    {
	if ((mi = (ModemInfoType *) l->data) == NULL)
	    continue;

	if (mi->name)
	    g_free(mi->name);
	if (mi->number)
	    g_free(mi->number);
	if (mi->parity)
	    g_free(mi->parity);
	if (mi->loginScript)
	    g_free(mi->loginScript);
	if (mi->logoutScript)
	    g_free(mi->logoutScript);
	if (mi->device)
	    g_free(mi->device);
	if (mi->initString)
	    g_free(mi->initString);
	g_free(mi);
    }
    if (ModemInfoList)
    {
	g_slist_free(ModemInfoList);
	ModemInfoList = NULL;
    }
}

/* ModemConnectByName() {{{1 */
int
ModemConnectByName(const char *name)
{
    GSList *slist;
    ModemInfoType *info = NULL;

    if (ModemInitialized)
    {
	StatusShowMessage(_("Currently connected. Close first..."));
	return FALSE;
    }

    g_return_val_if_fail(ModemInfo == NULL, -1);
    g_return_val_if_fail(Term != NULL, -1);
    g_return_val_if_fail(name != NULL, -1);
    g_return_val_if_fail(*name != 0, -1);

    Term->remoteWrite = ModemWrite;

    UseModem = TRUE;
    for (slist = ModemInfoList; slist; slist = slist->next)
    {
	info = slist->data;
	if (!strcasecmp(info->name, name) || !strcasecmp(info->number, name))
	    break;
    }
    if (!slist)	/* try with default setting */
    {
	if (ModemCheckDialNumber(name) != OK)
	{
	    StatusShowMessage(_("'%s' is not a valid dial number"), name);
	    return FALSE;
	}

	info = &DfltModemInfo;

	if (info->number)
	    g_free(info->number);

	info->name = "UNKNOWN";
	info->number = g_strdup(name);
	info->speed = DEFAULT_MODEM_SPEED;
	info->parity = DEFAULT_MODEM_PARITY;
	info->bits = DEFAULT_MODEM_BITS;
	info->useSWF = FALSE;
	info->useHWF = TRUE;
	info->loginScript = NULL;
	info->logoutScript = NULL;
	info->device = DEFAULT_MODEM_DEVICE;
	info->initString = DEFAULT_MODEM_INIT;
    }

    ModemInfo = info;

    if (InitModem(ModemInfo) == -1)
    {
	modemInfoFree();
	return FALSE;
    }

    TermReadInputTag = gdk_input_add(ReadFD, GDK_INPUT_READ,
				     TermInputHandler, NULL);
    return FALSE;
}

/* ModemHangup() {{{1 */
void
ModemHangup(void)
{
    if (ModemInitialized)
    {
	if (ModemDialInProgress)
	{
	    InputFilterList = g_slist_remove(InputFilterList,
					     ModemResponseWaitFilter);
	    ModemWrite(MODEM_DIAL_CANCEL_STR, strlen(MODEM_DIAL_CANCEL_STR));
	    gtk_timeout_remove(ModemDialTimeoutTag);
	    ModemDialTimeoutTag = 0;
	    ModemDialInProgress = FALSE;
	}
	if (ModemGetConnectionStatus())
	{
	    ModemDtrToggle(ReadFD);
	    ModemHangupInProgress = TRUE;	/* hack for direct connection */
	    ModemConnectionCheck(NULL);
	    ModemHangupInProgress = FALSE;
	}
	ModemRestore();
	modemHistoryDelete();
	ModemInitialized = FALSE;
    }
    if (ScriptRunning)
	ScriptCancel();
    if (LogFile)
	CloseLogFile(0);	/* safety */
    if (TermReadInputTag > 0)
    {
	gdk_input_remove(TermReadInputTag);
	TermReadInputTag = 0;
    }
    Term->remoteWrite = DummyFromTerm;

    modemInfoFree();
}
/* }}} */

/* MODEM USER COMMAND HISTORY */

static GtkWidget *modemHistWin = NULL;
static GtkWidget *modemHistList;

/* ModemOpenHistoryWin() {{{1 */
void
ModemOpenHistoryWin(void)
{
    GtkWidget *vbox, *swin, *sep, *hbox, *button;
    uint i, last;

    if (modemHistWin == NULL)
    {
	modemHistWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect(GTK_OBJECT(modemHistWin), "destroy",
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed),
			   &modemHistWin);
	gtk_window_set_title(GTK_WINDOW(modemHistWin), "Command History");
	gtk_container_set_border_width(GTK_CONTAINER(modemHistWin), 0);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(modemHistWin), vbox);

	swin = gtk_scrolled_window_new(NULL, NULL);
	gtk_container_set_border_width(GTK_CONTAINER(swin), 5);
	gtk_widget_set_usize(swin, -1, 200);
	gtk_box_pack_start(GTK_BOX(vbox), swin, TRUE, TRUE, 0);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swin),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_AUTOMATIC);

	modemHistList = gtk_clist_new(1);
	gtk_clist_set_selection_mode(GTK_CLIST(modemHistList),
				     GTK_SELECTION_EXTENDED);
	gtk_container_add(GTK_CONTAINER(swin), modemHistList);

	gtk_clist_freeze(GTK_CLIST(modemHistList));
	last = modemHistoryWrapped? ModemMaxHistory: modemHistoryIdx;
	for (i = 0; i < last; i++)
	    gtk_clist_append(GTK_CLIST(modemHistList), &modemHistoryBuf[i]);
	gtk_clist_thaw(GTK_CLIST(modemHistList));

	sep = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, TRUE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 0);

	button = gtk_button_new_with_label("Send+CR");
	gtk_container_set_border_width(GTK_CONTAINER(button), 5);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
				  GTK_SIGNAL_FUNC(modemHistorySendCR),
				  GTK_OBJECT(modemHistList));
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);

	button = gtk_button_new_with_label("Send");
	gtk_container_set_border_width(GTK_CONTAINER(button), 5);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
				  GTK_SIGNAL_FUNC(modemHistorySend),
				  GTK_OBJECT(modemHistList));
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);

	button = gtk_button_new_with_label("Close");
	gtk_container_set_border_width(GTK_CONTAINER(button), 5);
	gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
				  GTK_SIGNAL_FUNC(gtk_widget_destroy),
				  GTK_OBJECT(modemHistWin));
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);

	gtk_widget_show_all(modemHistWin);
    }
}

/* modemAddHistory() {{{1 */
static void
modemAddHistory(const guchar *s, int len)
{
    for (; len > 0; s++, len--)
    {
	int c = *s;
	if (c == '\n' || c == '\r')
	{
	    if (modemHistoryPos == 0)
		/* if user typed ENTER only */
		continue;

add_history_end:
	    modemHistoryBuf[modemHistoryIdx][modemHistoryPos] = '\0';
	    if (modemHistWin != NULL)
	    {
		gtk_clist_freeze(GTK_CLIST(modemHistList));
		if (modemHistoryWrapped)
		    gtk_clist_set_text(GTK_CLIST(modemHistList),
				       modemHistoryIdx, 0,
				       modemHistoryBuf[modemHistoryIdx]);
		else
		    gtk_clist_append(GTK_CLIST(modemHistList),
				     &modemHistoryBuf[modemHistoryIdx]);
		gtk_clist_thaw(GTK_CLIST(modemHistList));
	    }

	    if (++modemHistoryIdx == ModemMaxHistory)
	    {
		modemHistoryIdx = 0;
		modemHistoryWrapped = TRUE;
	    }
	    modemHistoryPos = 0;
	}
	else if (c == '\b')
	{
	    if (modemHistoryPos)
		modemHistoryPos--;
	}
	else
	{
	    modemHistoryBuf[modemHistoryIdx][modemHistoryPos] = c;
	    if (++modemHistoryPos == ModemHistoryBufSize)
	    {
		--modemHistoryPos;
		goto add_history_end;
	    }
	}
    }
}

/* modemHistoryRealSend() {{{1 */
static void
modemHistoryRealSend(GtkCList *clist, gboolean addCR)
{
    GList *t;

    gtk_clist_freeze(clist);

    for (t = clist->selection; t; t = t->next)
    {
	char *s;
	gint row = GPOINTER_TO_INT(t->data);
	if (gtk_clist_get_text(clist, row, 0, &s) == 0)
	{
	    g_warning("BUG: I believe it's bug!\n");
	    continue;
	}

	ModemWrite(s, strlen(s));
	if (addCR)
	{
	    char c = '\n';
	    ModemWrite(&c, 1);
	}

	/*   ʹ     ִ.
	 * target host Ҿ ɼ ũ,    ð
	 * ִ  ڴ. optionalϰ ó?
	 */
	PC_Sleep(300);
    }

    gtk_clist_thaw(clist);
}

/* modemHistorySendCR() {{{1 */
static void
modemHistorySendCR(GtkCList *clist)
{
    modemHistoryRealSend(clist, TRUE);
}

/* modemHistorySend() {{{1 */
static void
modemHistorySend(GtkCList *clist)
{
    modemHistoryRealSend(clist, FALSE);
}

/* ModemHistoryNew() {{{1 */
void
ModemHistoryNew(void)
{
    modemHistoryDelete();

    if (ModemMaxHistory && ModemHistoryBufSize)
    {
	modemHistoryBuf = (char **) malloc(ModemMaxHistory * sizeof(char *));
	if (modemHistoryBuf)
	{
	    guint32 i;

	    for (i = 0; i < ModemMaxHistory; i++)
	    {
		modemHistoryBuf[i] = (char *) malloc(ModemHistoryBufSize);
		if (modemHistoryBuf[i] == NULL)
		{
		    ModemMaxHistory = i;
		    if (ModemMaxHistory == 0)
		    {
			free(modemHistoryBuf);
			modemHistoryBuf = NULL;
		    }
		    g_warning("History buffer alloc failed.\n"
			      "Max available history number = %lu\n",
			      (unsigned long) ModemMaxHistory);
		    break;
		}
	    }
	}
    }
    else
	modemHistoryBuf = NULL;
    modemHistoryIdx = 0;
    modemHistoryPos = 0;
    modemHistoryWrapped = FALSE;
}

/* modemHistoryDelete() {{{1 */
static void
modemHistoryDelete(void)
{
    if (modemHistoryBuf)
    {
	guint32 i;

	for (i = 0; i < ModemMaxHistory; i++)
	    free(modemHistoryBuf[i]);
	free(modemHistoryBuf);
	modemHistoryBuf = NULL;
    }
}
